

class LifeLineChartCannotMoveIndividual(Exception):
    def __init__(self, *args):
        self.args = args


class LifeLineChartCollisionDetected(Exception):
    def __init__(self, *args):
        self.args = args


class LifeLineChartNotEnoughInformationToDisplay(Exception):
    def __init__(self, *args):
        self.args = args


class LifeLineChartUnknownPlacementError(Exception):
    def __init__(self, *args):
        self.args = args


class LifeLineChartUnknownSelectionAndConnectionError(Exception):
    def __init__(self, *args):
        self.args = args



"""
Life Line Chart Module
======================

This module can be used to generate genealogy charts.
"""

import logging

from .BaseChart import BaseChart
from .AncestorChart import AncestorChart
from .DescendantChart import DescendantChart

from .GraphicalIndividual import GraphicalIndividual
from .GraphicalFamily import GraphicalFamily

from .BaseFamily import BaseFamily
from .BaseIndividual import BaseIndividual, estimate_birth_date, estimate_death_date

from .InstanceContainer import InstanceContainer

# from .GedcomParsing import get_date_dict_from_tag, estimate_marriage_date, get_gedcom_instance_container

from .Exceptions import LifeLineChartCannotMoveIndividual, LifeLineChartCollisionDetected, LifeLineChartNotEnoughInformationToDisplay

logging.basicConfig()
logger = logging.getLogger("life_line_chart")
logger.setLevel(logging.FATAL)


from . import _version
__version__ = _version.get_versions()['version']



import names
import os
import datetime
from random import random


def generate_gedcom_file():
    """generate some gedcom file"""
    db = {}
    db['n_individuals'] = 0
    db['max_individuals'] = 8000
    db['n_families'] = 0
    db['yougest'] = None
    gedcom_content = """
    0 HEAD
1 SOUR Gramps
2 VERS 3.3.0
2 NAME Gramps
1 DATE {}
2 TIME 15:35:24
1 SUBM @SUBM@
1 COPR Copyright (c) 2020 Christian Schulze,,,.
1 GEDC
2 VERS 5.5
1 CHAR UTF-8
1 LANG German
""".format(datetime.date.today())

    def generate_individual(db, birth_year, sex=None, last_name=None):
        if not sex:
            sex = 'F' if random() < 0.5 else 'M'
        first_name = names.get_first_name(
            gender='male' if sex == 'M' else 'female')
        if random() < 0.3:
            first_name += ' ' + \
                names.get_first_name(gender='male' if sex == 'M' else 'female')
        if not last_name:
            last_name = names.get_last_name()
        birth_place = 'Paris' if random() < 0.5 else 'Rome'
        death_place = 'Zorge' if random() < 0.5 else 'Bruegge'
        db['n_individuals'] += 1
        individual_id = '@I{}@'.format(db["n_individuals"])
        death_year = birth_year + 40 + int(random()*20)
        db[individual_id] = {
            'birth': birth_year,
            'death': death_year,
            'sex': sex,
            'last_name': last_name
        }
        birth_date = '1 JUN {}'.format(birth_year)
        death_date = '1 JUN {}'.format(birth_year)
        if not db['yougest']:
            db['yougest'] = individual_id
        elif db[db['yougest']]['birth'] < birth_year:
            db['yougest'] = individual_id

        db[individual_id]['string'] = """0 {individual_id} INDI
1 NAME {first_name} /{last_name}/
1 SEX {sex}
1 BIRT
2 DATE {birth_date}
2 PLAC {birth_place}
1 DEAT
2 DATE {death_date}
2 PLAC {death_place}
""".format(**locals())
        return individual_id

    def generate_family(db, husband_id, wife_id, children_ids, marriage_year, marriage_place=None):
        if not marriage_place:
            marriage_place = 'London' if random() < 0.5 else 'Tokio'
        db['n_families'] += 1
        marriage_date = '1 MAY {}'.format(marriage_year)
        family_id = "@F{}@".format(db['n_families'])
        db[family_id] = {'string': """0 {family_id} FAM
1 HUSB {husband_id}
1 WIFE {wife_id}
1 MARR
2 DATE {marriage_date}
2 PLAC {marriage_place}
""".format(**locals())}
        for child_id in children_ids:
            db[family_id]['string'] += "1 CHIL {}\n".format(child_id)
        return family_id

    def find_by_birth_date(db, from_year, to_year, sex, exclude=[]):
        ids = []
        for individual_id, data in db.items():
            if not individual_id.startswith('@I'):
                continue
            if 'famc' in data:
                if data['birth'] > from_year and data['birth'] < to_year:
                    if sex == data['sex']:
                        if individual_id not in exclude:
                            ids.append(individual_id)
        if ids:
            return ids[int(random()*len(ids))]
        return None

    def generate_recursive_family(db, start_year=1000, generations=2, husband_id=None, wife_id=None, siblings=[], max_children=5):
        if not husband_id:
            if random() < 0.2:
                exclude = siblings.copy()
                if wife_id:
                    exclude += [wife_id]
                husband_id = find_by_birth_date(
                    db, start_year, start_year + 10, sex='M', exclude=exclude)
            if not husband_id:
                husband_id = generate_individual(
                    db, start_year + int(random()*5), sex='M')
            else:
                print('reused {}'.format(husband_id))
        if not wife_id:
            if random() < 10.9:
                exclude = siblings.copy() + [husband_id]
                wife_id = find_by_birth_date(
                    db, start_year, start_year + 10, sex='F', exclude=exclude)
            if not wife_id:
                wife_id = generate_individual(
                    db, start_year + int(random()*5), sex='F')
            else:
                print('reused {}'.format(wife_id))
        n_children = int((1+random()*(max_children-1)) *
                         (1 - db['n_individuals'] / db['max_individuals']))
        marriage_year = start_year + 20 + int(random()*5)
        children_ids = []
        for i in range(n_children):
            children_ids.append(generate_individual(
                db, birth_year=marriage_year + 1 + int(random()*10), last_name=db[husband_id]['last_name']))
        family_id = generate_family(
            db, husband_id, wife_id, children_ids, marriage_year)
        for i in range(n_children):
            db[children_ids[i]]['string'] += "1 FAMC "+family_id + '\n'
            db[children_ids[i]]['famc'] = family_id
            if generations > 0:
                generate_recursive_family(
                    db,
                    db[children_ids[i]]['birth'],
                    generations - 1,
                    children_ids[i] if db[children_ids[i]
                                          ]['sex'] == 'M' else None,
                    children_ids[i] if db[children_ids[i]
                                          ]['sex'] == 'F' else None,
                    children_ids)
        db[husband_id]['string'] += "1 FAMS "+family_id + '\n'
        db[wife_id]['string'] += "1 FAMS "+family_id + '\n'

    generate_recursive_family(db, generations=8, max_children=4)
    for k, v in db.items():
        if k.startswith('@I'):
            gedcom_content += v['string']
    for k, v in db.items():
        if k.startswith('@F'):
            gedcom_content += v['string']

    gedcom_content += '0 TRLR\n'
    open(os.path.join(os.path.dirname(__file__), '..', 'tests',
                      'autogenerated.ged'), 'w').write(gedcom_content)
# generate_gedcom_file()


def generate_individual_images():
    from PIL import Image, ImageDraw, ImageFont

    def generate_one_image(filename, text, font_size=22, pos=(15, 40), size=(100, 100), color=(160, 160, 160)):
        img = Image.new('RGB', size, color=color)

        d = ImageDraw.Draw(img)
        font = ImageFont.truetype(r'arial.ttf', font_size)
        d.text(pos, text, fill=(0, 0, 0), font=font)

        img.save(filename)

    for i in range(20):
        generate_one_image(
            'tests/images/individual_I6_image_age_{}.png'.format(
                1+i*4
            ), 'Age {}'.format(
                1+i*4,
            ))


generate_individual_images()



from copy import deepcopy
from collections import OrderedDict
import datetime


def estimate_death_date(individual):
    """
    if the death is unknown, then guess the death date by assuming the maximum age of 75 years or after 1900 100 years.

    Args:
        individual (BaseIndividual): individual instance
    """
    if individual.events['death_or_burial'] is None and not individual.events['birth_or_christening'] is None:
        date = individual.events['birth_or_christening']['date']
        max_age = 75
        if date.year > 1900:
            max_age = 100
        today = datetime.datetime.now()
        date_min = datetime.datetime(date.year+max_age-25, 12, 31)
        if date_min > today:
            date_min = today
        date_max = datetime.datetime(date.year+max_age, 12, 31)
        if date_max > today:
            date_max = today
        individual.events['death_or_burial'] = {
            'tag_name': 'None',
            'date': date_max,
            'ordinal_value': date_max.toordinal(),
            'ordinal_value_min': date_min.toordinal(),
            'ordinal_value_max': date_max.toordinal(),
            'comment': 'Estimated (max age)',
            'precision': 'y'
        }
        if date_max == today:
            individual.events['death_or_burial']['comment'] = 'Still alive'


def estimate_birth_date(individual, instances):
    """
    if the birth date is unknown, then estimate the birth date by assuming:
        - the birth took place at one year after the marriage
        - the individual was 25 on his/hers first marriage
        - the individual died with the maximum age of 75

    Args:
        individual (BaseIndividual): individual instance
        instances (InstanceContainer): instance container to get family information
    """
    if individual.events['birth_or_christening'] is None:
        # parents marriage
        for family_id in individual.child_of_family_id:
            parents_marriage = instances[('f', family_id)]
            if parents_marriage and parents_marriage.marriage:
                if (
                    not individual.events['birth_or_christening']
                    or individual.events['birth_or_christening']['ordinal_value'] < parents_marriage.marriage['ordinal_value']
                ):
                    if (
                        individual.events['birth_or_christening']
                        and individual.events['birth_or_christening']['ordinal_value'] < parents_marriage.marriage['ordinal_value']
                    ):
                        date = deepcopy(parents_marriage.marriage['date'])
                    else:
                        date = deepcopy(parents_marriage.marriage['date'])
                    date_min = datetime.datetime(
                        date.year+1, date.month, date.day, 0, 0, 0)
                    date_max = datetime.datetime(
                        date.year+11, date.month, date.day, 0, 0, 0)
                    individual.events['birth_or_christening'] = {
                        'tag_name': 'MARR',
                        'comment': 'Estimated (after parents marriage)',
                        'date': date_min,
                        'ordinal_value': date_min.toordinal(),
                        'ordinal_value_min': date_min.toordinal(),
                        'ordinal_value_max': date_max.toordinal(),
                        'precision': 'y'
                    }
    if individual.events['birth_or_christening'] is None:
        # at least 15 at marriage
        for marriage in individual.marriages:
            if marriage.marriage:
                if (
                    not individual.events['birth_or_christening']
                    or individual.events['birth_or_christening']['ordinal_value'] < marriage.marriage['ordinal_value']
                ):
                    if (
                        individual.events['birth_or_christening']
                        and individual.events['birth_or_christening']['ordinal_value'] < marriage.marriage['ordinal_value']
                    ):
                        date = deepcopy(marriage.marriage['date'])
                    else:
                        date = deepcopy(marriage.marriage['date'])
                    date_max = datetime.datetime(
                        date.year-20, date.month, date.day, 0, 0, 0)
                    date_min = datetime.datetime(
                        date.year-30, date.month, date.day, 0, 0, 0)
                    individual.events['birth_or_christening'] = {
                        'tag_name': 'MARR',
                        'comment': 'Estimated (min age at marriage)',
                        'date': date_min,
                        'ordinal_value': date_min.toordinal(),
                        'ordinal_value_min': date_min.toordinal(),
                        'ordinal_value_max': date_max.toordinal(),
                        'precision': 'y'
                    }
    if individual.events['birth_or_christening'] is None and not individual.events['death_or_burial'] is None:
        # max 75 years, so in birth can be estimated
        if 'death_or_burial' in individual.events:
            date = individual.events['death_or_burial']['date']
            max_age = 75
            if date.year > 1900:
                max_age = 100
            date_min = datetime.datetime(date.year-max_age, 1, 1)
            date_max = datetime.datetime(date.year-(max_age-25), 1, 1)
            individual.events['birth_or_christening'] = {
                'tag_name': 'None',
                'date': date_min,
                'ordinal_value': date_min.toordinal(),
                'ordinal_value_max': date_max.toordinal(),
                'ordinal_value_min': date_min.toordinal(),
                'comment': 'Estimated (max age)',
                'precision': 'y'
            }


class BaseIndividual():
    """
    Base class for individuals. This class is used as interface to the database.
    """
    # ordinal value of the birth date
    __birth_date_ov = None
    # ordinal value of the death date
    __death_date_ov = None

    def __init__(self, instances, individual_id):
        self._instances = instances
        self.individual_id = individual_id
        self._marriage_family_ids = []
        self.marriages = []
        self.child_of_family_id = []
        self.events = {}  # : events like birth or death
        self.graphical_representations = []  # : instances of graphical representations
        self.images = OrderedDict()  # : mapping of ordinal values to photos of this individual

    def __repr__(self):
        return 'individual "' + self.plain_name + '" ' + self.birth_date

    def _initialize(self):
        self._marriage_family_ids = self._get_marriage_family_ids()
        self.marriages = [
            self._instances[('f', m)] for m in self._marriage_family_ids
            if self._instances[('f', m)]]
        self.marriages.sort()

    def __lt__(self, other):
        """
        Sorting by birth date

        Args:
            other (BaseIndividual): the other instance

        Returns:
            bool: is less than
        """
        return self.birth_date_ov < other.birth_date_ov

    def get_name(self):
        if True:
            raise NotImplementedError()
        return ""

    @property
    def plain_name(self):
        return self._instances.display_plain_name(self)

    @property
    def children(self):
        """
        get the all children of this individual (all marriages)

        Returns:
            list: list of children individuals
        """
        children = []
        for m in self.marriages:
            children += m.children
        return children

    @property
    def birth_date(self):
        return self._instances.display_birth_date(self)

    @property
    def birth_date_ov(self):
        """
        get the ordinal value of the birth (or christening or baptism) date

        Returns:
            float: ordinal value of birth date
        """
        if self.__birth_date_ov is None:
            boc = self.events.get('birth_or_christening')
            if boc:
                self.__birth_date_ov = boc['ordinal_value']
                return self.__birth_date_ov
            return None
        else:
            return self.__birth_date_ov

    @property
    def birth_label(self):
        """
        get the birth label used for displaying

        Returns:
            str: birth label
        """
        string = ''
        if self.events['birth_or_christening']:
            date_str = self.birth_date
            event = self.events['birth_or_christening']
            if event['comment']:
                string = self._instances.date_label_translation[event['comment']].format(
                    symbol='*', date=date_str)
            else:
                string += '*\xa0' + date_str
        return string

    @property
    def death_label(self):
        """
        get death label used for displaying

        Returns:
            str: death label
        """
        string = ''
        if self.events['death_or_burial']:
            date_str = self.death_date
            event = self.events['death_or_burial']
            if event['comment']:
                string = self._instances.date_label_translation[event['comment']].format(
                    symbol='\u2020', date=date_str)
            else:
                string += '\u2020\xa0' + date_str
        return string

    @property
    def death_date(self):
        return self._instances.display_death_date(self)

    @property
    def death_date_ov(self):
        """
        get the ordinal value of the death (or burial) date

        Returns:
            float: ordinal value of death date
        """
        if self.__death_date_ov is None:
            dob = self.events.get('death_or_burial')
            if dob:
                self.__death_date_ov = dob['ordinal_value']
                return self.__death_date_ov
            return None
        else:
            return self.__death_date_ov

    @property
    def info_text(self):
        return self._instances.display_info_text(self)

    @property
    def short_info_text(self):
        return self._instances.display_short_info_text(self)

    def has_marriages(self):
        return len(self._marriage_family_ids) > 0

    @property
    def child_of_families(self):
        families = [
            self._instances[('f', family_id)] for family_id in self.child_of_family_id
            if self._instances[('f', family_id)]]
        return families

    def has_graphical_representation(self):
        return len(self.graphical_representations) > 0

    def _get_marriage_family_ids(self):
        if True:
            raise NotImplementedError()
        return []



try:
    from collections.abc import MutableSequence
except ImportError:
    # Required for python versions < 3.9
    from collections import MutableSequence


class Line(object):
    def __init__(self, start, end):
        self.start = start
        self.end = end

    def __repr__(self):
        return 'Line(start=%s, end=%s)' % (self.start, self.end)

    def __eq__(self, other):
        if not isinstance(other, Line):
            return NotImplemented
        return self.start == other.start and self.end == other.end

    def __ne__(self, other):
        if not isinstance(other, Line):
            return NotImplemented
        return not self == other

    def __getitem__(self, item):
        return self.bpoints()[item]

    def __len__(self):
        return 2

    # def joins_smoothly_with(self, previous, wrt_parameterization=False):
    #     """Checks if this segment joins smoothly with previous segment.  By
    #     default, this only checks that this segment starts moving (at t=0) in
    #     the same direction (and from the same positive) as previous stopped
    #     moving (at t=1).  To check if the tangent magnitudes also match, set
    #     wrt_parameterization=True."""
    #     if wrt_parameterization:
    #         return self.start == previous.end and np.isclose(
    #             self.derivative(0), previous.derivative(1))
    #     else:
    #         return self.start == previous.end and np.isclose(
    #             self.unit_tangent(0), previous.unit_tangent(1))

    def point(self, t):
        """returns the coordinates of the Bezier curve evaluated at t."""
        distance = self.end - self.start
        return self.start + distance*t

    # def length(self, t0=0, t1=1, error=None, min_depth=None):
    #     """returns the length of the line segment between t0 and t1."""
    #     return abs(self.end - self.start)*(t1-t0)

    # def ilength(self, s, s_tol=ILENGTH_S_TOL, maxits=ILENGTH_MAXITS,
    #             error=ILENGTH_ERROR, min_depth=ILENGTH_MIN_DEPTH):
    #     """Returns a float, t, such that self.length(0, t) is approximately s.
    #     See the inv_arclength() docstring for more details."""
    #     return inv_arclength(self, s, s_tol=s_tol, maxits=maxits, error=error,
    #                          min_depth=min_depth)

    def bpoints(self):
        """returns the Bezier control points of the segment."""
        return self.start, self.end

    def poly(self, return_coeffs=False):
        """returns the line as a Polynomial object."""
        p = self.bpoints()
        coeffs = ([p[1] - p[0], p[0]])
    #     if return_coeffs:
    #         return coeffs
    #     else:
    #         return np.poly1d(coeffs)
        return coeffs

    # def derivative(self, t=None, n=1):
    #     """returns the nth derivative of the segment at t."""
    #     assert self.end != self.start
    #     if n == 1:
    #         return self.end - self.start
    #     elif n > 1:
    #         return 0
    #     else:
    #         raise ValueError("n should be a positive integer.")

    # def unit_tangent(self, t=None):
    #     """returns the unit tangent of the segment at t."""
    #     assert self.end != self.start
    #     dseg = self.end - self.start
    #     return dseg/abs(dseg)

    # def normal(self, t=None):
    #     """returns the (right hand rule) unit normal vector to self at t."""
    #     return -1j*self.unit_tangent(t)

    # def curvature(self, t):
    #     """returns the curvature of the line, which is always zero."""
    #     return 0

    # # def icurvature(self, kappa):
    # #     """returns a list of t-values such that 0 <= t<= 1 and
    # #     seg.curvature(t) = kappa."""
    # #     if kappa:
    # #         raise ValueError("The .icurvature() method for Line elements will "
    # #                          "return an empty list if kappa is nonzero and "
    # #                          "will raise this exception when kappa is zero as "
    # #                          "this is true at every point on the line.")
    # #     return []

    # def reversed(self):
    #     """returns a copy of the Line object with its orientation reversed."""
    #     return Line(self.end, self.start)

    # def intersect(self, other_seg, tol=None):
    #     """Finds the intersections of two segments.
    #     returns a list of tuples (t1, t2) such that
    #     self.point(t1) == other_seg.point(t2).
    #     Note: This will fail if the two segments coincide for more than a
    #     finite collection of points.
    #     tol is not used."""
    #     if isinstance(other_seg, Line):
    #         assert other_seg.end != other_seg.start and self.end != self.start
    #         assert self != other_seg
    #         # Solve the system [p1-p0, q1-q0]*[t1, t2]^T = q0 - p0
    #         # where self == Line(p0, p1) and other_seg == Line(q0, q1)
    #         a = (self.start.real, self.end.real)
    #         b = (self.start.imag, self.end.imag)
    #         c = (other_seg.start.real, other_seg.end.real)
    #         d = (other_seg.start.imag, other_seg.end.imag)
    #         denom = ((a[1] - a[0])*(d[0] - d[1]) -
    #                  (b[1] - b[0])*(c[0] - c[1]))
    #         if np.isclose(denom, 0):
    #             return []
    #         t1 = (c[0]*(b[0] - d[1]) -
    #               c[1]*(b[0] - d[0]) -
    #               a[0]*(d[0] - d[1]))/denom
    #         t2 = -(a[1]*(b[0] - d[0]) -
    #                a[0]*(b[1] - d[0]) -
    #                c[0]*(b[0] - b[1]))/denom
    #         if 0 <= t1 <= 1 and 0 <= t2 <= 1:
    #             return [(t1, t2)]
    #         return []
    #     elif isinstance(other_seg, QuadraticBezier):
    #         t2t1s = bezier_by_line_intersections(other_seg, self)
    #         return [(t1, t2) for t2, t1 in t2t1s]
    #     elif isinstance(other_seg, CubicBezier):
    #         t2t1s = bezier_by_line_intersections(other_seg, self)
    #         return [(t1, t2) for t2, t1 in t2t1s]
    #     elif isinstance(other_seg, Arc):
    #         t2t1s = other_seg.intersect(self)
    #         return [(t1, t2) for t2, t1 in t2t1s]
    #     elif isinstance(other_seg, Path):
    #         raise TypeError(
    #             "other_seg must be a path segment, not a Path object, use "
    #             "Path.intersect().")
    #     else:
    #         raise TypeError("other_seg must be a path segment.")

    # def bbox(self):
    #     """returns the bounding box for the segment in the form
    #     (xmin, xmax, ymin, ymax)."""
    #     xmin = min(self.start.real, self.end.real)
    #     xmax = max(self.start.real, self.end.real)
    #     ymin = min(self.start.imag, self.end.imag)
    #     ymax = max(self.start.imag, self.end.imag)
    #     return xmin, xmax, ymin, ymax

    # def cropped(self, t0, t1):
    #     """returns a cropped copy of this segment which starts at
    #     self.point(t0) and ends at self.point(t1)."""
    #     return Line(self.point(t0), self.point(t1))

    # def split(self, t):
    #     """returns two segments, whose union is this segment and which join at
    #     self.point(t)."""
    #     pt = self.point(t)
    #     return Line(self.start, pt), Line(pt, self.end)

    # def radialrange(self, origin, return_all_global_extrema=False):
    #     """returns the tuples (d_min, t_min) and (d_max, t_max) which minimize
    #     and maximize, respectively, the distance d = |self.point(t)-origin|."""
    #     return bezier_radialrange(self, origin,
    #             return_all_global_extrema=return_all_global_extrema)

    # def rotated(self, degs, origin=None):
    #     """Returns a copy of self rotated by `degs` degrees (CCW) around the
    #     point `origin` (a complex number).  By default `origin` is either
    #     `self.point(0.5)`, or in the case that self is an Arc object,
    #     `origin` defaults to `self.center`."""
    #     return rotate(self, degs, origin=origin)

    # def translated(self, z0):
    #     """Returns a copy of self shifted by the complex quantity `z0` such
    #     that self.translated(z0).point(t) = self.point(t) + z0 for any t."""
    #     return translate(self, z0)

    # def scaled(self, sx, sy=None, origin=0j):
    #     """Scale transform.  See `scale` function for further explanation."""
    #     return scale(self, sx=sx, sy=sy, origin=origin)


class CubicBezier(object):
    # For compatibility with old pickle files.
    # _length_info = {'length': None, 'bpoints': None, 'error': None,
    #                 'min_depth': None}

    def __init__(self, start, control1, control2, end):
        self.start = start
        self.control1 = control1
        self.control2 = control2
        self.end = end

        # used to know if self._length needs to be updated
        # self._length_info = {'length': None, 'bpoints': None, 'error': None,
        #                      'min_depth': None}

    def __repr__(self):
        return 'CubicBezier(start=%s, control1=%s, control2=%s, end=%s)' % (
            self.start, self.control1, self.control2, self.end)

    def __eq__(self, other):
        if not isinstance(other, CubicBezier):
            return NotImplemented
        return self.start == other.start and self.end == other.end \
            and self.control1 == other.control1 \
            and self.control2 == other.control2

    def __ne__(self, other):
        if not isinstance(other, CubicBezier):
            return NotImplemented
        return not self == other

    def __getitem__(self, item):
        return self.bpoints()[item]

    def __len__(self):
        return 4

    # def is_smooth_from(self, previous, warning_on=True):
    #     """[Warning: The name of this method is somewhat misleading (yet kept
    #     for compatibility with scripts created using svg.path 2.0).  This
    #     method is meant only for d string creation and should not be used to
    #     check for kinks.  To check a segment for differentiability, use the
    #     joins_smoothly_with() method instead.]"""
    #     if warning_on:
    #         warn(_is_smooth_from_warning)
    #     if isinstance(previous, CubicBezier):
    #         return (self.start == previous.end and
    #                 (self.control1 - self.start) == (
    #                     previous.end - previous.control2))
    #     else:
    #         return self.control1 == self.start

    # def joins_smoothly_with(self, previous, wrt_parameterization=False):
    #     """Checks if this segment joins smoothly with previous segment.  By
    #     default, this only checks that this segment starts moving (at t=0) in
    #     the same direction (and from the same positive) as previous stopped
    #     moving (at t=1).  To check if the tangent magnitudes also match, set
    #     wrt_parameterization=True."""
    #     if wrt_parameterization:
    #         return self.start == previous.end and np.isclose(
    #             self.derivative(0), previous.derivative(1))
    #     else:
    #         return self.start == previous.end and np.isclose(
    #             self.unit_tangent(0), previous.unit_tangent(1))

    def point(self, t):
        """Evaluate the cubic Bezier curve at t using Horner's rule."""
        # algebraically equivalent to
        # P0*(1-t)**3 + 3*P1*t*(1-t)**2 + 3*P2*(1-t)*t**2 + P3*t**3
        # for (P0, P1, P2, P3) = self.bpoints()
        return self.start + t*(
            3*(self.control1 - self.start) + t*(
                3*(self.start + self.control2) - 6*self.control1 + t*(
                    -self.start + 3*(self.control1 - self.control2) + self.end
                )))

    # def length(self, t0=0, t1=1, error=LENGTH_ERROR, min_depth=LENGTH_MIN_DEPTH):
    #     """Calculate the length of the path up to a certain position"""
    #     if t0 == 0 and t1 == 1:
    #         if self._length_info['bpoints'] == self.bpoints() \
    #                 and self._length_info['error'] >= error \
    #                 and self._length_info['min_depth'] >= min_depth:
    #             return self._length_info['length']

    #     # using scipy.integrate.quad is quick
    #     if _quad_available:
    #         s = quad(lambda tau: abs(self.derivative(tau)), t0, t1,
    #                         epsabs=error, limit=1000)[0]
    #     else:
    #         s = segment_length(self, t0, t1, self.point(t0), self.point(t1),
    #                            error, min_depth, 0)

    #     if t0 == 0 and t1 == 1:
    #         self._length_info['length'] = s
    #         self._length_info['bpoints'] = self.bpoints()
    #         self._length_info['error'] = error
    #         self._length_info['min_depth'] = min_depth
    #         return self._length_info['length']
    #     else:
    #         return s

    # def ilength(self, s, s_tol=ILENGTH_S_TOL, maxits=ILENGTH_MAXITS,
    #             error=ILENGTH_ERROR, min_depth=ILENGTH_MIN_DEPTH):
    #     """Returns a float, t, such that self.length(0, t) is approximately s.
    #     See the inv_arclength() docstring for more details."""
    #     return inv_arclength(self, s, s_tol=s_tol, maxits=maxits, error=error,
    #                          min_depth=min_depth)

    def bpoints(self):
        """returns the Bezier control points of the segment."""
        return self.start, self.control1, self.control2, self.end

    def poly(self, return_coeffs=False):
        """Returns a the cubic as a Polynomial object."""
        p = self.bpoints()
        coeffs = (-p[0] + 3*(p[1] - p[2]) + p[3],
                  3*(p[0] - 2*p[1] + p[2]),
                  3*(-p[0] + p[1]),
                  p[0])
        # if return_coeffs:
        #     return coeffs
        # else:
        #     return np.poly1d(coeffs)
        return coeffs

    # def derivative(self, t, n=1):
    #     """returns the nth derivative of the segment at t.
    #     Note: Bezier curves can have points where their derivative vanishes.
    #     If you are interested in the tangent direction, use the unit_tangent()
    #     method instead."""
    #     p = self.bpoints()
    #     if n == 1:
    #         return 3*(p[1] - p[0])*(1 - t)**2 + 6*(p[2] - p[1])*(1 - t)*t + 3*(
    #             p[3] - p[2])*t**2
    #     elif n == 2:
    #         return 6*(
    #             (1 - t)*(p[2] - 2*p[1] + p[0]) + t*(p[3] - 2*p[2] + p[1]))
    #     elif n == 3:
    #         return 6*(p[3] - 3*(p[2] - p[1]) - p[0])
    #     elif n > 3:
    #         return 0
    #     else:
    #         raise ValueError("n should be a positive integer.")

    # def unit_tangent(self, t):
    #     """returns the unit tangent vector of the segment at t (centered at
    #     the origin and expressed as a complex number).  If the tangent
    #     vector's magnitude is zero, this method will find the limit of
    #     self.derivative(tau)/abs(self.derivative(tau)) as tau approaches t."""
    #     return bezier_unit_tangent(self, t)

    # def normal(self, t):
    #     """returns the (right hand rule) unit normal vector to self at t."""
    #     return -1j * self.unit_tangent(t)

    # def curvature(self, t):
    #     """returns the curvature of the segment at t."""
    #     return segment_curvature(self, t)

    # # def icurvature(self, kappa):
    # #     """returns a list of t-values such that 0 <= t<= 1 and
    # #     seg.curvature(t) = kappa."""
    # #     z = self.poly()
    # #     x, y = real(z), imag(z)
    # #     dx, dy = x.deriv(), y.deriv()
    # #     ddx, ddy = dx.deriv(), dy.deriv()
    # #
    # #     p = kappa**2*(dx**2 + dy**2)**3 - (dx*ddy - ddx*dy)**2
    # #     return polyroots01(p)

    # def reversed(self):
    #     """returns a copy of the CubicBezier object with its orientation
    #     reversed."""
    #     new_cub = CubicBezier(self.end, self.control2, self.control1,
    #                           self.start)
    #     if self._length_info['length']:
    #         new_cub._length_info = self._length_info
    #         new_cub._length_info['bpoints'] = (
    #             self.end, self.control2, self.control1, self.start)
    #     return new_cub

    # def intersect(self, other_seg, tol=1e-12):
    #     """Finds the intersections of two segments.
    #     returns a list of tuples (t1, t2) such that
    #     self.point(t1) == other_seg.point(t2).
    #     Note: This will fail if the two segments coincide for more than a
    #     finite collection of points."""
    #     if isinstance(other_seg, Line):
    #         return bezier_by_line_intersections(self, other_seg)
    #     elif (isinstance(other_seg, QuadraticBezier) or
    #           isinstance(other_seg, CubicBezier)):
    #         assert self != other_seg
    #         longer_length = max(self.length(), other_seg.length())
    #         return bezier_intersections(self, other_seg,
    #                                     longer_length=longer_length,
    #                                     tol=tol, tol_deC=tol)
    #     elif isinstance(other_seg, Arc):
    #         t2t1s = other_seg.intersect(self)
    #         return [(t1, t2) for t2, t1 in t2t1s]
    #     elif isinstance(other_seg, Path):
    #         raise TypeError(
    #             "other_seg must be a path segment, not a Path object, use "
    #             "Path.intersect().")
    #     else:
    #         raise TypeError("other_seg must be a path segment.")

    # def bbox(self):
    #     """returns the bounding box for the segment in the form
    #     (xmin, xmax, ymin, ymax)."""
    #     return bezier_bounding_box(self)

    # def split(self, t):
    #     """returns two segments, whose union is this segment and which join at
    #     self.point(t)."""
    #     bpoints1, bpoints2 = split_bezier(self.bpoints(), t)
    #     return CubicBezier(*bpoints1), CubicBezier(*bpoints2)

    # def cropped(self, t0, t1):
    #     """returns a cropped copy of this segment which starts at
    #     self.point(t0) and ends at self.point(t1)."""
    #     return CubicBezier(*crop_bezier(self, t0, t1))

    # def radialrange(self, origin, return_all_global_extrema=False):
    #     """returns the tuples (d_min, t_min) and (d_max, t_max) which minimize
    #     and maximize, respectively, the distance d = |self.point(t)-origin|."""
    #     return bezier_radialrange(self, origin,
    #             return_all_global_extrema=return_all_global_extrema)

    # def rotated(self, degs, origin=None):
    #     """Returns a copy of self rotated by `degs` degrees (CCW) around the
    #     point `origin` (a complex number).  By default `origin` is either
    #     `self.point(0.5)`, or in the case that self is an Arc object,
    #     `origin` defaults to `self.center`."""
    #     return rotate(self, degs, origin=origin)

    # def translated(self, z0):
    #     """Returns a copy of self shifted by the complex quantity `z0` such
    #     that self.translated(z0).point(t) = self.point(t) + z0 for any t."""
    #     return translate(self, z0)

    # def scaled(self, sx, sy=None, origin=0j):
    #     """Scale transform.  See `scale` function for further explanation."""
    #     return scale(self, sx=sx, sy=sy, origin=origin)


class Path(MutableSequence):
    """A Path is a sequence of path segments"""

    # Put it here, so there is a default if unpickled.
    _closed = False
    _start = None
    _end = None

    def __init__(self, *segments, **kw):
        self._segments = list(segments)
        self._length = None
        self._lengths = None
        if 'closed' in kw:
            self.closed = kw['closed']  # DEPRECATED
        if self._segments:
            self._start = self._segments[0].start
            self._end = self._segments[-1].end
        else:
            self._start = None
            self._end = None

    def __getitem__(self, index):
        return self._segments[index]

    def __setitem__(self, index, value):
        self._segments[index] = value
        self._length = None
        self._start = self._segments[0].start
        self._end = self._segments[-1].end

    def __delitem__(self, index):
        del self._segments[index]
        self._length = None
        self._start = self._segments[0].start
        self._end = self._segments[-1].end

    def __iter__(self):
        return self._segments.__iter__()

    def __contains__(self, x):
        return self._segments.__contains__(x)

    def insert(self, index, value):
        self._segments.insert(index, value)
        self._length = None
        self._start = self._segments[0].start
        self._end = self._segments[-1].end

    def reversed(self):
        """returns a copy of the Path object with its orientation reversed."""
        newpath = [seg.reversed() for seg in self]
        newpath.reverse()
        return Path(*newpath)

    def __len__(self):
        return len(self._segments)

    def __repr__(self):
        return "Path({})".format(
            ",\n     ".join(repr(x) for x in self._segments))

    def __eq__(self, other):
        if not isinstance(other, Path):
            return NotImplemented
        if len(self) != len(other):
            return False
        for s, o in zip(self._segments, other._segments):
            if not s == o:
                return False
        return True

    def __ne__(self, other):
        if not isinstance(other, Path):
            return NotImplemented
        return not self == other

    # def _calc_lengths(self, error=LENGTH_ERROR, min_depth=LENGTH_MIN_DEPTH):
    #     if self._length is not None:
    #         return

    #     lengths = [each.length(error=error, min_depth=min_depth) for each in
    #                self._segments]
    #     self._length = sum(lengths)
    #     self._lengths = [each/self._length for each in lengths]

    # def point(self, pos):

    #     # Shortcuts
    #     if pos == 0.0:
    #         return self._segments[0].point(pos)
    #     if pos == 1.0:
    #         return self._segments[-1].point(pos)

    #     self._calc_lengths()
    #     # Find which segment the point we search for is located on:
    #     segment_start = 0
    #     for index, segment in enumerate(self._segments):
    #         segment_end = segment_start + self._lengths[index]
    #         if segment_end >= pos:
    #             # This is the segment! How far in on the segment is the point?
    #             segment_pos = (pos - segment_start)/(
    #                 segment_end - segment_start)
    #             return segment.point(segment_pos)
    #         segment_start = segment_end

    # def length(self, T0=0, T1=1, error=LENGTH_ERROR, min_depth=LENGTH_MIN_DEPTH):
    #     self._calc_lengths(error=error, min_depth=min_depth)
    #     if T0 == 0 and T1 == 1:
    #         return self._length
    #     else:
    #         if len(self) == 1:
    #             return self[0].length(t0=T0, t1=T1)
    #         idx0, t0 = self.T2t(T0)
    #         idx1, t1 = self.T2t(T1)
    #         if idx0 == idx1:
    #             return self[idx0].length(t0=t0, t1=t1)
    #         return (self[idx0].length(t0=t0) +
    #                 sum(self[idx].length() for idx in range(idx0 + 1, idx1)) +
    #                 self[idx1].length(t1=t1))

    # def ilength(self, s, s_tol=ILENGTH_S_TOL, maxits=ILENGTH_MAXITS,
    #             error=ILENGTH_ERROR, min_depth=ILENGTH_MIN_DEPTH):
    #     """Returns a float, t, such that self.length(0, t) is approximately s.
    #     See the inv_arclength() docstring for more details."""
    #     return inv_arclength(self, s, s_tol=s_tol, maxits=maxits, error=error,
    #                          min_depth=min_depth)

    # def iscontinuous(self):
    #     """Checks if a path is continuous with respect to its
    #     parameterization."""
    #     return all(self[i].end == self[i+1].start for i in range(len(self) - 1))

    # def continuous_subpaths(self):
    #     """Breaks self into its continuous components, returning a list of
    #     continuous subpaths.
    #     I.e.
    #     (all(subpath.iscontinuous() for subpath in self.continuous_subpaths())
    #      and self == concatpaths(self.continuous_subpaths()))
    #     )
    #     """
    #     subpaths = []
    #     subpath_start = 0
    #     for i in range(len(self) - 1):
    #         if self[i].end != self[(i+1) % len(self)].start:
    #             subpaths.append(Path(*self[subpath_start: i+1]))
    #             subpath_start = i+1
    #     subpaths.append(Path(*self[subpath_start: len(self)]))
    #     return subpaths

    # def isclosed(self):
    #     """This function determines if a connected path is closed."""
    #     assert len(self) != 0
    #     assert self.iscontinuous()
    #     return self.start == self.end

    # def isclosedac(self):
    #     assert len(self) != 0
    #     return self.start == self.end

    # def _is_closable(self):
    #     end = self[-1].end
    #     for segment in self:
    #         if segment.start == end:
    #             return True
    #     return False

    # @property
    # def closed(self, warning_on=CLOSED_WARNING_ON):
    #     """The closed attribute is deprecated, please use the isclosed()
    #     method instead.  See _closed_warning for more information."""
    #     mes = ("This attribute is deprecated, consider using isclosed() "
    #            "method instead.\n\nThis attribute is kept for compatibility "
    #            "with scripts created using svg.path (v2.0). You can prevent "
    #            "this warning in the future by setting "
    #            "CLOSED_WARNING_ON=False.")
    #     if warning_on:
    #         warn(mes)
    #     return self._closed and self._is_closable()

    # @closed.setter
    # def closed(self, value):
    #     value = bool(value)
    #     if value and not self._is_closable():
    #         raise ValueError("End does not coincide with a segment start.")
    #     self._closed = value

    # @property
    # def start(self):
    #     if not self._start:
    #         self._start = self._segments[0].start
    #     return self._start

    # @start.setter
    # def start(self, pt):
    #     self._start = pt
    #     self._segments[0].start = pt

    # @property
    # def end(self):
    #     if not self._end:
    #         self._end = self._segments[-1].end
    #     return self._end

    # @end.setter
    # def end(self, pt):
    #     self._end = pt
    #     self._segments[-1].end = pt

    def d(self, useSandT=False, use_closed_attrib=False):
        """Returns a path d-string for the path object.
        For an explanation of useSandT and use_closed_attrib, see the
        compatibility notes in the README."""

        if use_closed_attrib:
            self_closed = self.closed(warning_on=False)
            if self_closed:
                segments = self[:-1]
            else:
                segments = self[:]
        else:
            self_closed = False
            segments = self[:]

        current_pos = None
        parts = []
        previous_segment = None
        end = self[-1].end

        for segment in segments:
            seg_start = segment.start
            # If the start of this segment does not coincide with the end of
            # the last segment or if this segment is actually the close point
            # of a closed path, then we should start a new subpath here.
            if current_pos != seg_start or \
                    (self_closed and seg_start == end and use_closed_attrib):
                parts.append('M {},{}'.format(seg_start.real, seg_start.imag))

            if isinstance(segment, Line):
                args = segment.end.real, segment.end.imag
                parts.append('L {},{}'.format(*args))
            elif isinstance(segment, CubicBezier):
                if useSandT and segment.is_smooth_from(previous_segment,
                                                       warning_on=False):
                    args = (segment.control2.real, segment.control2.imag,
                            segment.end.real, segment.end.imag)
                    parts.append('S {},{} {},{}'.format(*args))
                else:
                    args = (segment.control1.real, segment.control1.imag,
                            segment.control2.real, segment.control2.imag,
                            segment.end.real, segment.end.imag)
                    parts.append('C {},{} {},{} {},{}'.format(*args))
            # elif isinstance(segment, QuadraticBezier):
            #     if useSandT and segment.is_smooth_from(previous_segment,
            #                                            warning_on=False):
            #         args = segment.end.real, segment.end.imag
            #         parts.append('T {},{}'.format(*args))
            #     else:
            #         args = (segment.control.real, segment.control.imag,
            #                 segment.end.real, segment.end.imag)
            #         parts.append('Q {},{} {},{}'.format(*args))

            # elif isinstance(segment, Arc):
            #     args = (segment.radius.real, segment.radius.imag,
            #             segment.rotation,int(segment.large_arc),
            #             int(segment.sweep),segment.end.real, segment.end.imag)
            #     parts.append('A {},{} {} {:d},{:d} {},{}'.format(*args))
            current_pos = segment.end
            previous_segment = segment

        if self_closed:
            parts.append('Z')

        return ' '.join(parts)

    # def joins_smoothly_with(self, previous, wrt_parameterization=False):
    #     """Checks if this Path object joins smoothly with previous
    #     path/segment.  By default, this only checks that this Path starts
    #     moving (at t=0) in the same direction (and from the same positive) as
    #     previous stopped moving (at t=1).  To check if the tangent magnitudes
    #     also match, set wrt_parameterization=True."""
    #     if wrt_parameterization:
    #         return self[0].start == previous.end and self.derivative(
    #             0) == previous.derivative(1)
    #     else:
    #         return self[0].start == previous.end and self.unit_tangent(
    #             0) == previous.unit_tangent(1)

    # def T2t(self, T):
    #     """returns the segment index, `seg_idx`, and segment parameter, `t`,
    #     corresponding to the path parameter `T`.  In other words, this is the
    #     inverse of the `Path.t2T()` method."""
    #     if T == 1:
    #         return len(self)-1, 1
    #     if T == 0:
    #         return 0, 0
    #     self._calc_lengths()
    #     # Find which segment self.point(T) falls on:
    #     T0 = 0  # the T-value the current segment starts on
    #     for seg_idx, seg_length in enumerate(self._lengths):
    #         T1 = T0 + seg_length  # the T-value the current segment ends on
    #         if T1 >= T:
    #             # This is the segment!
    #             t = (T - T0)/seg_length
    #             return seg_idx, t
    #         T0 = T1

    #     assert 0 <= T <= 1
    #     raise BugException

    # def t2T(self, seg, t):
    #     """returns the path parameter T which corresponds to the segment
    #     parameter t.  In other words, for any Path object, path, and any
    #     segment in path, seg,  T(t) = path.t2T(seg, t) is the unique
    #     reparameterization such that path.point(T(t)) == seg.point(t) for all
    #     0 <= t <= 1.
    #     Input Note: seg can be a segment in the Path object or its
    #     corresponding index."""
    #     self._calc_lengths()
    #     # Accept an index or a segment for seg
    #     if isinstance(seg, int):
    #         seg_idx = seg
    #     else:
    #         try:
    #             seg_idx = self.index(seg)
    #         except ValueError:
    #             assert is_path_segment(seg) or isinstance(seg, int)
    #             raise

    #     segment_start = sum(self._lengths[:seg_idx])
    #     segment_end = segment_start + self._lengths[seg_idx]
    #     T = (segment_end - segment_start)*t + segment_start
    #     return T

    # def derivative(self, T, n=1):
    #     """returns the tangent vector of the Path at T (centered at the origin
    #     and expressed as a complex number).
    #     Note: Bezier curves can have points where their derivative vanishes.
    #     If you are interested in the tangent direction, use unit_tangent()
    #     method instead."""
    #     seg_idx, t = self.T2t(T)
    #     seg = self._segments[seg_idx]
    #     return seg.derivative(t, n=n)/seg.length()**n

    # def unit_tangent(self, T):
    #     """returns the unit tangent vector of the Path at T (centered at the
    #     origin and expressed as a complex number).  If the tangent vector's
    #     magnitude is zero, this method will find the limit of
    #     self.derivative(tau)/abs(self.derivative(tau)) as tau approaches T."""
    #     seg_idx, t = self.T2t(T)
    #     return self._segments[seg_idx].unit_tangent(t)

    # def normal(self, t):
    #     """returns the (right hand rule) unit normal vector to self at t."""
    #     return -1j*self.unit_tangent(t)

    # def curvature(self, T):
    #     """returns the curvature of this Path object at T and outputs
    #     float('inf') if not differentiable at T."""
    #     seg_idx, t = self.T2t(T)
    #     seg = self[seg_idx]
    #     if np.isclose(t, 0) and (seg_idx != 0 or self.end==self.start):
    #         previous_seg_in_path = self._segments[
    #             (seg_idx - 1) % len(self._segments)]
    #         if not seg.joins_smoothly_with(previous_seg_in_path):
    #             return float('inf')
    #     elif np.isclose(t, 1) and (seg_idx != len(self) - 1 or self.end==self.start):
    #         next_seg_in_path = self._segments[
    #             (seg_idx + 1) % len(self._segments)]
    #         if not next_seg_in_path.joins_smoothly_with(seg):
    #             return float('inf')
    #     dz = self.derivative(T)
    #     ddz = self.derivative(T, n=2)
    #     dx, dy = dz.real, dz.imag
    #     ddx, ddy = ddz.real, ddz.imag
    #     return abs(dx*ddy - dy*ddx)/(dx*dx + dy*dy)**1.5

    # # def icurvature(self, kappa):
    # #     """returns a list of T-values such that 0 <= T <= 1 and
    # #     seg.curvature(t) = kappa.
    # #     Note: not implemented for paths containing Arc segments."""
    # #     assert is_bezier_path(self)
    # #     Ts = []
    # #     for i, seg in enumerate(self):
    # #         Ts += [self.t2T(i, t) for t in seg.icurvature(kappa)]
    # #     return Ts

    # def area(self):
    #     """returns the area enclosed by this Path object.
    #     Note: negative area results from CW (as opposed to CCW)
    #     parameterization of the Path object."""
    #     assert self.isclosed()
    #     area_enclosed = 0
    #     for seg in self:
    #         x = real(seg.poly())
    #         dy = imag(seg.poly()).deriv()
    #         integrand = x*dy
    #         integral = integrand.integ()
    #         area_enclosed += integral(1) - integral(0)
    #     return area_enclosed

    # def intersect(self, other_curve, justonemode=False, tol=1e-12):
    #     """returns list of pairs of pairs ((T1, seg1, t1), (T2, seg2, t2))
    #     giving the intersection points.
    #     If justonemode==True, then returns just the first
    #     intersection found.
    #     tol is used to check for redundant intersections (see comment above
    #     the code block where tol is used).
    #     Note:  If the two path objects coincide for more than a finite set of
    #     points, this code will fail."""
    #     path1 = self
    #     if isinstance(other_curve, Path):
    #         path2 = other_curve
    #     else:
    #         path2 = Path(other_curve)
    #     assert path1 != path2
    #     intersection_list = []
    #     for seg1 in path1:
    #         for seg2 in path2:
    #             if justonemode and intersection_list:
    #                 return intersection_list[0]
    #             for t1, t2 in seg1.intersect(seg2, tol=tol):
    #                 T1 = path1.t2T(seg1, t1)
    #                 T2 = path2.t2T(seg2, t2)
    #                 intersection_list.append(((T1, seg1, t1), (T2, seg2, t2)))
    #     if justonemode and intersection_list:
    #         return intersection_list[0]

    #     # Note: If the intersection takes place at a joint (point one seg ends
    #     # and next begins in path) then intersection_list may contain a
    #     # redundant intersection.  This code block checks for and removes said
    #     # redundancies.
    #     if intersection_list:
    #         pts = [seg1.point(_t1) for _T1, _seg1, _t1 in list(zip(*intersection_list))[0]]
    #         indices2remove = []
    #         for ind1 in range(len(pts)):
    #             for ind2 in range(ind1 + 1, len(pts)):
    #                 if abs(pts[ind1] - pts[ind2]) < tol:
    #                     # then there's a redundancy. Remove it.
    #                     indices2remove.append(ind2)
    #         intersection_list = [inter for ind, inter in
    #                              enumerate(intersection_list) if
    #                              ind not in indices2remove]
    #     return intersection_list

    # def bbox(self):
    #     """returns a bounding box for the input Path object in the form
    #     (xmin, xmax, ymin, ymax)."""
    #     bbs = [seg.bbox() for seg in self._segments]
    #     xmins, xmaxs, ymins, ymaxs = list(zip(*bbs))
    #     xmin = min(xmins)
    #     xmax = max(xmaxs)
    #     ymin = min(ymins)
    #     ymax = max(ymaxs)
    #     return xmin, xmax, ymin, ymax

    # def cropped(self, T0, T1):
    #     """returns a cropped copy of the path."""
    #     assert 0 <= T0 <= 1 and 0 <= T1<= 1
    #     assert T0 != T1
    #     assert not (T0 == 1 and T1 == 0)

    #     if T0 == 1 and 0 < T1 < 1 and self.isclosed():
    #         return self.cropped(0, T1)

    #     if T1 == 1:
    #         seg1 = self[-1]
    #         t_seg1 = 1
    #         i1 = len(self) - 1
    #     else:
    #         seg1_idx, t_seg1 = self.T2t(T1)
    #         seg1 = self[seg1_idx]
    #         if np.isclose(t_seg1, 0):
    #             i1 = (self.index(seg1) - 1) % len(self)
    #             seg1 = self[i1]
    #             t_seg1 = 1
    #         else:
    #             i1 = self.index(seg1)
    #     if T0 == 0:
    #         seg0 = self[0]
    #         t_seg0 = 0
    #         i0 = 0
    #     else:
    #         seg0_idx, t_seg0 = self.T2t(T0)
    #         seg0 = self[seg0_idx]
    #         if np.isclose(t_seg0, 1):
    #             i0 = (self.index(seg0) + 1) % len(self)
    #             seg0 = self[i0]
    #             t_seg0 = 0
    #         else:
    #             i0 = self.index(seg0)

    #     if T0 < T1 and i0 == i1:
    #         new_path = Path(seg0.cropped(t_seg0, t_seg1))
    #     else:
    #         new_path = Path(seg0.cropped(t_seg0, 1))

    #         # T1<T0 must cross discontinuity case
    #         if T1 < T0:
    #             if not self.isclosed():
    #                 raise ValueError("This path is not closed, thus T0 must "
    #                                  "be less than T1.")
    #             else:
    #                 for i in range(i0 + 1, len(self)):
    #                     new_path.append(self[i])
    #                 for i in range(0, i1):
    #                     new_path.append(self[i])

    #         # T0<T1 straight-forward case
    #         else:
    #             for i in range(i0 + 1, i1):
    #                 new_path.append(self[i])

    #         if t_seg1 != 0:
    #             new_path.append(seg1.cropped(0, t_seg1))
    #     return new_path

    # def radialrange(self, origin, return_all_global_extrema=False):
    #     """returns the tuples (d_min, t_min, idx_min), (d_max, t_max, idx_max)
    #     which minimize and maximize, respectively, the distance
    #     d = |self[idx].point(t)-origin|."""
    #     if return_all_global_extrema:
    #         raise NotImplementedError
    #     else:
    #         global_min = (np.inf, None, None)
    #         global_max = (0, None, None)
    #         for seg_idx, seg in enumerate(self):
    #             seg_global_min, seg_global_max = seg.radialrange(origin)
    #             if seg_global_min[0] < global_min[0]:
    #                 global_min = seg_global_min + (seg_idx,)
    #             if seg_global_max[0] > global_max[0]:
    #                 global_max = seg_global_max + (seg_idx,)
    #         return global_min, global_max

    # def rotated(self, degs, origin=None):
    #     """Returns a copy of self rotated by `degs` degrees (CCW) around the
    #     point `origin` (a complex number).  By default `origin` is either
    #     `self.point(0.5)`, or in the case that self is an Arc object,
    #     `origin` defaults to `self.center`."""
    #     return rotate(self, degs, origin=origin)

    # def translated(self, z0):
    #     """Returns a copy of self shifted by the complex quantity `z0` such
    #     that self.translated(z0).point(t) = self.point(t) + z0 for any t."""
    #     return translate(self, z0)

    # def scaled(self, sx, sy=None, origin=0j):
    #     """Scale transform.  See `scale` function for further explanation."""
    #     return scale(self, sx=sx, sy=sy, origin=origin)



import logging
from copy import deepcopy
from collections import OrderedDict
from .BaseSVGChart import BaseSVGChart
from .Translation import get_strings, recursive_merge_dict_members

logger = logging.getLogger("life_line_chart")

_strings = get_strings('DescendantChart')
_, _strings = recursive_merge_dict_members(BaseSVGChart.SETTINGS_DESCRIPTION, _strings, remove_unknown_keys=False)


class DescendantChart(BaseSVGChart):
    """
    Descendant Chart
    ================

    The descendant chart shows the descendants of one or more root individuals.
    The parents enclose direct children. Both, father and mother are visible.

    Each individual appears once. So in case of a second marriage, the
    individual is connected across the chart to the second spouse. Because
    of that, pedigree collapse can be visualized.
    """

    DEFAULT_FORMATTING = {
        'highlight_descendants': False,
    }
    DEFAULT_FORMATTING.update(BaseSVGChart.DEFAULT_FORMATTING)

    DEFAULT_POSITIONING = {
        'chart_layout': 'enclosing',
    }
    DEFAULT_POSITIONING.update(BaseSVGChart.DEFAULT_POSITIONING)

    DEFAULT_CHART_CONFIGURATION = {
        'root_individuals': [],
        'discovery_blacklist': []
    }
    DEFAULT_CHART_CONFIGURATION.update(BaseSVGChart.DEFAULT_CHART_CONFIGURATION)

    SETTINGS_DESCRIPTION = _strings

    def __init__(self, positioning=None, formatting=None, instance_container=None):
        BaseSVGChart.__init__(self, positioning, formatting, instance_container)

        # configuration of this chart
        self._chart_configuration.update(DescendantChart.DEFAULT_CHART_CONFIGURATION)

    def select_descendants(self, individual, gr_child_of_family, generations=None, filter=None):
        """
        Create graphical representations for all descendants.

        Args:
            individual (BaseIndividual): parent individual
            gr_child_of_family (GraphicalFamily): parent family
            generations (int, optional): number of generations to go deeper. Defaults to None.
            filter (lambda, optional): filter for individuals. Defaults to None.
        """
        if individual is None or filter and filter(individual):
            return

        gr_individual = self._create_individual_graphical_representation(
            individual, not self._positioning['unique_graphical_representation'])

        if gr_individual is None:
            return

        if gr_child_of_family is None and individual.child_of_families:
            gr_child_of_family = self._create_family_graphical_representation(
                individual.child_of_families[0], not self._positioning['unique_graphical_representation'])

        for marriage in individual.marriages:
            if self._positioning['chart_layout'] == 'enclosing':
                if marriage.has_graphical_representation() and self._positioning['unique_graphical_representation']:
                    continue

            if generations > 0 or generations < 0:
                new_gr_marriage = not marriage.has_graphical_representation()
                gr_marriage = self._create_family_graphical_representation(
                    marriage, not self._positioning['unique_graphical_representation'])

                if gr_marriage.husb == individual:
                    gr_marriage.gr_husb = gr_individual
                else:
                    gr_marriage.gr_wife = gr_individual

                if self._positioning['chart_layout'] == 'enclosing':
                    spouse = marriage.get_spouse(individual.individual_id)
                    if spouse is not None:
                        if filter is None or not filter(spouse):
                            gr_spouse = self._create_individual_graphical_representation(
                                spouse, not self._positioning['unique_graphical_representation'])

                            if gr_marriage.husb == spouse:
                                gr_marriage.gr_husb = gr_spouse
                            else:
                                gr_marriage.gr_wife = gr_spouse

                if new_gr_marriage or not self._positioning['unique_graphical_representation']:
                    for child in marriage.children:
                        gr_child = self.select_descendants(
                            child, gr_marriage, generations - 1, filter=filter)
                        if gr_child:
                            gr_marriage.add_visible_children(gr_child)
                            # gr_child.ancestor_chart_parent_family_placement = gr_marriage
                    cofs = individual.child_of_families
                    for cof in cofs[:1]:
                        gr_marriage.visual_placement_parent_family = gr_child_of_family
                        gr_marriage.descendant_chart_parent_family_placement = gr_child_of_family
        return gr_individual

    def place_selected_individuals_cactus(self, gr_individual, gr_child_of_family, x_offset=0, x_offset_root=None, discovery_cache=None):
        """
        Place the graphical representations in direction of x.

        Args:
            gr_individual (GraphicalIndividual): individual
            gr_child_of_family (GraphicalFamily): child-of-family of this individual
            x_offset (int): starting position
            discovery_cache (list): list of discovered individuals
        """
        if discovery_cache is None:
            discovery_cache = []
        individual = gr_individual.individual
        discovery_cache.append(gr_individual)
        logger.info("discovering {}".format(individual.plain_name))

        x_position = x_offset
        self.min_x_index = min(self.min_x_index, x_position)

        # marriages which have been placed over this parent family
        visible_local_marriages = \
            [marriage for marriage in gr_individual.visible_marriages
                if gr_child_of_family is None or
             marriage.descendant_chart_parent_family_placement == gr_child_of_family]

        # get childrens widths
        child_widths = OrderedDict()
        total_number_of_descendants = 0
        for gr_marriage in reversed(visible_local_marriages):
            for gr_child in gr_marriage.visible_children:
                gr_child_descendants = gr_child.get_all_descendants()
                child_widths[gr_child.g_id] = gr_child_descendants
                total_number_of_descendants += len(gr_child_descendants) + 1

        gr_individual.special_properties['number_of_descendants'] = total_number_of_descendants

        # get root position
        number_of_placed_descendants = 0
        split_children = [OrderedDict(), OrderedDict()]
        for gr_marriage in visible_local_marriages:
            split_children[0][gr_marriage.g_id] = []
            split_children[1][gr_marriage.g_id] = []
            vcs = gr_marriage.visible_children
            sorted_vcs = sorted(vcs, key=lambda t: len(child_widths[t.g_id]))
            reordered_vcs = []
            for index in range(len(sorted_vcs)):
                if index*2 < len(sorted_vcs):
                    reordered_vcs.append(sorted_vcs[index*2])
                else:
                    index2 = index*2 - len(sorted_vcs)
                    if len(sorted_vcs) % 2 == 0:
                        index2 += 1
                    reordered_vcs.append(sorted_vcs[index2])
                # sorted_vcs[(2*i) % len(sorted_vcs)]
            for gr_child in reordered_vcs:
                this_step = len(child_widths[gr_child.g_id]) + 1
                if max(2, total_number_of_descendants) / 2 + 0.5 >= number_of_placed_descendants + this_step:
                    number_of_placed_descendants += this_step
                    split_children[0][gr_marriage.g_id].append(gr_child)
                else:
                    split_children[1][gr_marriage.g_id].append(gr_child)

        root_individual_position = x_position + number_of_placed_descendants

        individual_has_been_placed = False
        number_of_placed_descendants = 0
        for split_index in range(2):
            # for marriage_index, gr_marriage in enumerate(reversed(visible_local_marriages)):
            #     vcs = split_children[split_index][marriage_index].copy()
            #     sorted_vcs = sorted(vcs, key=lambda t: t.birth_date_ov if split_index == 0 else - t.birth_date_ov)
            #     for gr_child in sorted_vcs:
            #         self.place_selected_individuals_cactus(
            #             gr_child, gr_marriage, x_position, root_individual_position,
            #             discovery_cache=discovery_cache)
            #         width = len(child_widths[gr_child.g_id]) + 1
            #         # width = gr_child.get_descendant_width(
            #         #     gr_marriage)
            #         x_position += width
            #         number_of_placed_descendants += width
            vcs = []
            for gr_marriage in visible_local_marriages:
                c = split_children[split_index][gr_marriage.g_id]
                vcs += zip(c, [gr_marriage]*len(c))
            if split_index == 0:
                sorted_vcs = sorted(vcs, key=lambda t: t[0].birth_date_ov)
            else:
                sorted_vcs = sorted(vcs, key=lambda t: -t[0].birth_date_ov)
            for gr_child, gr_marriage in sorted_vcs:
                self.place_selected_individuals_cactus(
                    gr_child, gr_marriage, x_position, root_individual_position,
                    discovery_cache=discovery_cache)
                width = len(child_widths[gr_child.g_id]) + 1
                # width = gr_child.get_descendant_width(
                #     gr_marriage)
                x_position += width
                number_of_placed_descendants += width
            if not individual_has_been_placed and len(visible_local_marriages):
                individual_has_been_placed = True
                if not gr_individual.has_position_vector(gr_child_of_family):
                    # x_offset_root = x_offset
                    if x_offset_root is None:
                        x_offset_root = root_individual_position
                    gr_individual.set_position_vector(
                        x_offset_root, gr_child_of_family, True, overrule_ov=1)
                    for gr_marriage in gr_individual.visible_marriages:
                        if not gr_individual.has_position_vector(gr_marriage):
                            gr_individual.set_position_vector(
                                x_position, gr_marriage)
                    x_position += 1

        if not individual_has_been_placed:
            # x_offset_root = x_offset
            if x_offset_root is None:
                x_offset_root = root_individual_position
            gr_individual.set_position_vector(
                x_offset_root, gr_child_of_family, True, overrule_ov=1)
            for gr_marriage in gr_individual.visible_marriages:
                gr_individual.set_position_vector(
                    x_position, gr_marriage)
            if not gr_individual.visible_marriages:
                gr_individual.set_position_vector(
                    x_position, None)
            x_position += 1

        self.max_x_index = max(self.max_x_index, x_position)

        # recalculate
        birth_ordinal_value = gr_individual.birth_date_ov
        death_ordinal_value = gr_individual.death_date_ov
        if self.min_ordinal is not None and self.max_ordinal is not None:
            self.min_ordinal = min(self.min_ordinal, birth_ordinal_value)
            self.max_ordinal = max(self.max_ordinal, death_ordinal_value)
        elif death_ordinal_value and birth_ordinal_value:
            self.min_ordinal = birth_ordinal_value
            self.max_ordinal = death_ordinal_value

    def place_selected_individuals_enclosing(self, gr_individual, gr_child_of_family, x_offset=0, discovery_cache=[]):
        """
        Place the graphical representations in direction of x.

        Args:
            gr_individual (GraphicalIndividual): individual
            gr_child_of_family (GraphicalFamily): child-of-family of this individual
            x_offset (int): starting position
            discovery_cache (list): list of discovered individuals
        """
        individual = gr_individual.individual
        discovery_cache.append(individual.plain_name)
        logger.info("discovering {}".format(individual.plain_name))

        x_position = x_offset
        self.min_x_index = min(self.min_x_index, x_position)

        # marriages which have been placed over this parent family
        visible_local_marriages = [
            marriage for marriage in gr_individual.visible_marriages
            if gr_child_of_family is None or
            marriage.descendant_chart_parent_family_placement == gr_child_of_family]

        if len(visible_local_marriages) == 0:
            gr_individual.set_position_vector(
                x_position, gr_child_of_family, True)
            x_position += 1

        total_number_of_descendants = len(gr_individual.get_all_descendants())
        gr_individual.special_properties['number_of_descendants'] = total_number_of_descendants

        for marriage_index, gr_marriage in enumerate(reversed(visible_local_marriages)):
            gr_spouse = gr_marriage.get_gr_spouse(gr_individual)

            if gr_spouse:
                total_number_of_descendants = len(gr_spouse.get_all_descendants())
                gr_spouse.special_properties['number_of_descendants'] = total_number_of_descendants

            # starting x index of gr_individual is first marriage (i.e. last in reversed list)
            if marriage_index == len(visible_local_marriages) - 1:
                if not gr_individual.has_position_vector(gr_child_of_family):
                    gr_individual.set_position_vector(
                        x_position, gr_child_of_family, True)

            if marriage_index == len(visible_local_marriages) - 1:
                if not gr_individual.has_position_vector(gr_marriage):
                    gr_individual.set_position_vector(
                        x_position, gr_marriage)
                    x_position += 1
            else:
                if gr_spouse and not gr_spouse.has_position_vector(gr_marriage):
                    gr_spouse.set_position_vector(
                        x_position, gr_marriage)
                    x_position += 1

            for gr_child in gr_marriage.visible_children:
                self.place_selected_individuals_enclosing(
                    gr_child, gr_marriage, x_position,
                    discovery_cache=discovery_cache)
                width = gr_child.get_descendant_width(
                    gr_marriage)
                x_position += width

            if marriage_index < len(visible_local_marriages) - 1:
                if not gr_individual.has_position_vector(gr_marriage):
                    gr_individual.set_position_vector(
                        x_position, gr_marriage)
                    x_position += 1
            else:
                if gr_spouse and not gr_spouse.has_position_vector(gr_marriage):
                    gr_spouse.set_position_vector(
                        x_position, gr_marriage)
                    x_position += 1

        self.max_x_index = max(self.max_x_index, x_position)

        # recalculate
        birth_ordinal_value = gr_individual.birth_date_ov
        death_ordinal_value = gr_individual.death_date_ov
        if self.min_ordinal is not None and self.max_ordinal is not None:
            self.min_ordinal = min(self.min_ordinal, birth_ordinal_value)
            self.max_ordinal = max(self.max_ordinal, death_ordinal_value)
        elif death_ordinal_value and birth_ordinal_value:
            self.min_ordinal = birth_ordinal_value
            self.max_ordinal = death_ordinal_value

    def modify_layout(self, root_individual_id):
        """
        Improvement of individual placement.

        Args:
            root_individual_id (str): root individual id used as root node for compression
        """
        self.check_unique_x_position(False)
        self._check_compressed_x_position(
            False, self.position_to_person_map)

    def update_chart(
        self, filter_lambda=None, color_lambda=None,
        images_lambda=None, rebuild_all=False, update_view=False,
        clear_before_rebuild=True
    ):
        """
        Update the chart, caching of positioning data is regarded

        Args:
            filter_lambda (lambda(BaseIndividual), optional): filtering of individuals. Defaults to None.
            color_lambda (lambda(GraphicalIndividual), optional): coloring of individuals. Defaults to None.
            images_lambda (lambda(BaseIndividual), optional): images of individuals. Defaults to None.
            rebuild_all (bool, optional): clear cache, rebuild everything. Defaults to False.
            update_view (bool, optional): update formatting only. Defaults to False.
            clear_before_rebuild (bool, optional): clear instances before rebuilding the chart. Defaults to True.

        Returns:
            bool: view has changed
        """
        rebuild_all = rebuild_all or self._positioning != self._backup_positioning or \
            self._chart_configuration != self._backup_chart_configuration
        update_view = update_view or rebuild_all or self._formatting != self._backup_formatting

        def local_filter_lambda(individual, _filter_lambda=filter_lambda):
            if individual.individual_id in self._chart_configuration['discovery_blacklist']:
                return True
            if _filter_lambda is not None:
                return _filter_lambda(individual)
            return False

        if color_lambda:
            update_view = True
        if images_lambda:
            update_view = True
        self._instances.color_getter = self._instances.color_getters[self._formatting['coloring_of_individuals']]

        def weighting_lambda(gr_individual):
            if self._formatting['line_weighting'] == 'number_of_descendants':
                if 'number_of_descendants' in gir.special_properties:
                    return 1.5 - 1/(0.3*gir.special_properties['number_of_descendants']+1)
            return 1

        if rebuild_all:
            if clear_before_rebuild:
                self.clear_graphical_representations()
            for settings in self._chart_configuration['root_individuals']:
                root_individual_id = settings['individual_id']
                generations = settings['generations']
                root_individual = self._instances[(
                    'i', root_individual_id)]
                self.select_descendants(root_individual, None, generations, filter=local_filter_lambda)

            x_pos = 0
            for settings in self._chart_configuration['root_individuals']:
                if 'x_offset' in settings:
                    x_pos += settings['x_offset']
                root_individual_id = settings['individual_id']
                generations = settings['generations']
                root_individual = self._instances[(
                    'i', root_individual_id)]
                if root_individual is None:
                    continue
                gr_root_individual = root_individual.graphical_representations[0]
                cof_family_id = None
                if root_individual.child_of_family_id:
                    cof_family_id = root_individual.child_of_family_id[0]
                cof_family = self._instances[('f', cof_family_id)]
                gr_cof_family = None
                if cof_family:
                    gr_cof_family = cof_family.graphical_representations[0]
                if self._positioning['chart_layout'] == 'cactus':
                    self.place_selected_individuals_cactus(
                        gr_root_individual, gr_cof_family, x_pos)
                else:
                    self.place_selected_individuals_enclosing(
                        gr_root_individual, gr_cof_family, x_pos)

                x_pos += gr_root_individual.get_descendant_width(gr_cof_family)

            for settings in self._chart_configuration['root_individuals']:
                root_individual_id = settings['individual_id']
                generations = settings['generations']
                try:
                    self.modify_layout(root_individual_id)
                except Exception:
                    pass

            for gir in self.gr_individuals:
                gir.weight = weighting_lambda(gir)
                color = None
                if color_lambda:
                    color = color_lambda(gir)
                if color_lambda is None or color is None:
                    if self._formatting['highlight_descendants']:
                        cofs = gir.individual.child_of_families
                        if not cofs or not cofs[0].has_graphical_representation():
                            marriages = gir.visible_marriages
                            if marriages:
                                gr_spouse = marriages[0].get_gr_spouse(gir)
                                if (
                                    gr_spouse and gr_spouse.individual.child_of_families
                                    and gr_spouse.individual.child_of_families[0].has_graphical_representation()
                                ):
                                    color = (225, 225, 225)
                    if color is None:
                        color = self._instances.color_getter(gir)
                gir.color = color
                if images_lambda:
                    images = images_lambda(gir.individual)
                else:
                    images = OrderedDict()
                gir.individual.images = images

            self.define_svg_items()

        elif update_view:
            self.clear_svg_items()

            for gir in self.gr_individuals:
                gir.weight = weighting_lambda(gir)
                color = None
                if color_lambda:
                    color = color_lambda(gir)
                if color_lambda is None or color is None:
                    if self._formatting['highlight_descendants']:
                        cofs = gir.individual.child_of_families
                        if not cofs or not cofs[0].has_graphical_representation():
                            marriages = gir.visible_marriages
                            if marriages:
                                spouse = marriages[0].get_spouse(gir)
                                if (
                                    spouse and spouse.individual.child_of_families
                                    and spouse.individual.child_of_families[0].has_graphical_representation()
                                ):
                                    color = (225, 225, 225)
                    if color is None:
                        color = self._instances.color_getter(gir)
                gir.color = color
                if images_lambda:
                    images = images_lambda(gir.individual)
                else:
                    images = OrderedDict()
                gir.individual.images = images

            self.define_svg_items()
        self._backup_chart_configuration = deepcopy(self._chart_configuration)
        self._backup_formatting = deepcopy(self._formatting)
        self._backup_positioning = deepcopy(self._positioning)
        return update_view or rebuild_all



from .GedcomIndividual import GedcomIndividual
from .Exceptions import LifeLineChartUnknownPlacementError
from collections import OrderedDict


class GraphicalIndividual():
    """
    Class which represents one appearance of an individual

    If a parent family is strongly connected to a child, then the family will
    be moved if the child is moved. Weak connections are ignored.

    There are two degrees of freedom when placing the parent family.
    1. If several siblings are visible, then one sibling needs to be selected
    for the strong connection.
    2. If the selected sibling was married several times, then the marriage
    needs to be selected.
    """

    # x positions in different family appearances
    _x_position = None
    # color of this individual
    color = [200, 200, 255]
    # ordinal value of the birth date
    __birth_date_ov = None
    # ordinal value of the death date
    __death_date_ov = None
    weight = 1

    def __init__(self, instances, individual_id):
        self.items = []
        self.individual_id = individual_id
        self.__instances = instances
        self.individual = self.__instances[('i', self.individual_id)]
        self.individual.graphical_representations.append(self)
        self.debug_label = ""
        self.g_id = None

        # in ancestor charts children can be added, which are strongly connected, but
        # since they are a dead end, not a root, they cannot be used for placement
        self.qualified_for_placement = True

        self.special_properties = {}

        self.ancestor_chart_parent_family_placement = None, None

    def __repr__(self):
        return 'gr_individual "' + self.individual.plain_name + '" ' + self.individual.birth_date

    def __lt__(self, other):
        """
        Sorting by birth date

        Args:
            other (GraphicalIndividual): the other instance

        Returns:
            bool: is less than
        """
        return self.birth_date_ov < other.birth_date_ov

    def __eq__(self, other):
        """
        == operator

        Args:
            other (GraphicalIndividual): other instance

        Returns:
            bool: is equal
        """
        if type(other) != GraphicalIndividual:
            return False
        return self.g_id == other.g_id

    def get_marriages(self):
        marriages = self.individual.marriages
        return [m.graphical_representations[0] for m in marriages if m.has_graphical_representation()]

    def get_ancestor_width(self, gr_family):
        """
        Width of the ancestor individuals which are strongly connected.
        (only tested with ancestor chart)

        Args:
            gr_family (GraphicalFamily): gr_family which is examined

        Returns:
            int: width
        """
        x_min, x_max = self.get_ancestor_range(gr_family)
        width = x_max - x_min + 1
        return width

    def get_ancestor_range(self, gr_family):
        """
        Get the x range from min to max.
        (only tested with ancestor chart)

        Args:
            gr_family (GraphicalFamily): family which is examined

        Returns:
            tuple: x_min, x_max
        """
        gr_family_g_id = None
        if gr_family is not None:
            gr_family_g_id = gr_family.g_id
            # at least root node has None
        if (self.g_id, gr_family_g_id) in self.__instances.ancestor_width_cache:
            # caching
            return self.__instances.ancestor_width_cache[(self.g_id, gr_family_g_id)]
        x_v = [self._x_position[gr_family_g_id][1]]
        x_min = x_v.copy()
        x_max = x_v.copy()
        # if [3] is true, then that index is the ancestor family

        strongly_connected_parent_family, strongly_connected_spouse_family = self.ancestor_chart_parent_family_placement

        if strongly_connected_parent_family and strongly_connected_spouse_family == gr_family:
            gr_father = strongly_connected_parent_family.gr_husb
            if gr_father:
                # only handle if the father is visible
                f_x_min, f_x_max = gr_father.get_ancestor_range(
                    strongly_connected_parent_family)
                x_min.append(f_x_min)
                x_max.append(f_x_max)
            gr_mother = strongly_connected_parent_family.gr_wife
            if gr_mother:
                # only handle if the mother is visible
                m_x_min, m_x_max = gr_mother.get_ancestor_range(
                    strongly_connected_parent_family)
                x_min.append(m_x_min)
                x_max.append(m_x_max)
            # add siblings
            x_v = [gr_c.get_x_index(strongly_connected_parent_family.g_id) for gr_c in strongly_connected_parent_family.visible_children
                   if strongly_connected_parent_family.g_id in gr_c.get_position_dict()]
            x_min += x_v
            x_max += x_v

        x_min = min(x_min)
        x_max = max(x_max)
        self.__instances.ancestor_width_cache[(self.g_id, gr_family_g_id)] = x_min, x_max
        return x_min, x_max

    def get_descendant_width(self, gr_family):
        """
        Width of the descendant individuals which are strongly connected.
        (only tested with descendant chart)

        Args:
            gr_family (GraphicalFamily): family which is examined

        Returns:
            int: width
        """
        x_min, x_max = self.get_descendant_range(gr_family)
        width = x_max - x_min + 1
        return width

    def get_descendant_range(self, gr_family):
        """
        Get the x range from min to max.
        (only tested with descendant chart)

        Args:
            gr_family (GraphicalFamily): family which is examined

        Returns:
            tuple: x_min, x_max
        """
        family_id = None
        family = None
        family_g_id = None
        if gr_family is not None:
            family_id = gr_family.family_id
            family = gr_family.family
            family_g_id = gr_family.g_id
            # at least root node has None
        if (self.individual_id, family_id) in self.__instances.ancestor_width_cache:
            # caching
            # return self.__instances.ancestor_width_cache[(self.individual_id, family_id)]
            pass

        x_min = []
        x_max = []

        parent_family = self.__instances[('f', family_id)]
        if parent_family is None or not parent_family.has_graphical_representation():
            # return 1
            pass

        # marriages which have been placed over this parent family
        visible_local_marriages = [
            marriage for marriage in self.visible_marriages
            if gr_family is None or
            marriage.descendant_chart_parent_family_placement == gr_family
        ]

        marriages = self.individual.marriages
        for vm in visible_local_marriages:
            marriage = vm.family
            gr_marriage = vm
            if (family_id is None or gr_marriage.visual_placement_parent_family is not None and
                gr_marriage.visual_placement_parent_family.family_id == family_id
                ):
                x_min.append(self._x_position[gr_marriage.g_id][1])
                x_max.append(self._x_position[gr_marriage.g_id][1])

                for gr_child in gr_marriage.visible_children:
                    c_x_min, c_x_max = gr_child.get_descendant_range(gr_marriage)
                    x_min.append(c_x_min)
                    x_max.append(c_x_max)

                gr_spouse = gr_marriage.get_gr_spouse(self)
                if gr_spouse:
                    x_v = gr_spouse.get_x_index(gr_marriage.g_id)
                    x_min.append(x_v)
                    x_max.append(x_v)

        if len(x_min) == 0 and len(x_max) == 0:
            x_v = [self._x_position[family_g_id][1]]
            x_min += x_v
            x_max += x_v
        x_min = min(x_min)
        x_max = max(x_max)
        self.__instances.ancestor_width_cache[(self.individual_id, family_id)] = x_min, x_max
        return x_min, x_max

    @property
    def visible_marriages(self):
        """
        List of marriages which have a graphical representation.

        Returns:
            list: list of visible marriages
        """
        if self.g_id not in self.__instances.connection_container['i']:
            return []
        connected_parent_families = []
        for g_id, connections in self.__instances.connection_container['i'][self.g_id].items():
            if 'gr_husb' in connections or 'gr_wife' in connections:
                connected_parent_families.append(self.__instances[('f', g_id[1])].graphical_representations[g_id[0]])
        if len(connected_parent_families) > 0:
            connected_parent_families.sort()
            return connected_parent_families
        return []

    @property
    def connected_parent_families(self):
        """
        Child of families where this individual appears.

        Returns:
            list: list of parent families
        """
        if self.g_id not in self.__instances.connection_container['i']:
            return []
        connected_parent_families = []
        for g_id, connections in self.__instances.connection_container['i'][self.g_id].items():
            if 'weak_child' in connections:
                connected_parent_families.append(self.__instances[('f', g_id[1])].graphical_representations[g_id[0]])
        if len(connected_parent_families) > 0:
            return connected_parent_families
        return []

    @property
    def ancestor_chart_parent_family_placement(self):
        """
        This is the tuple with parent family and spouse/marriage family which are strongly connected.
        The parent family is placed above the individual at the marriage family position.

        Raises:
            LifeLineChartUnknownPlacementError: placing error

        Returns:
            tuple: strongly connected parent family and spouse family
        """
        if self.g_id not in self.__instances.connection_container['i']:
            return None
        strongly_connected_parent_families = []
        strongly_connected_spouse_families = []
        for g_id, connections in self.__instances.connection_container['i'][self.g_id].items():
            if 'strong_child' in connections:
                strongly_connected_parent_families.append(self.__instances[('f', g_id[1])].graphical_representations[g_id[0]])
            if 'strong_marriage' in connections:
                strongly_connected_spouse_families.append(self.__instances[('f', g_id[1])].graphical_representations[g_id[0]])
        if len(strongly_connected_parent_families) > 1:
            raise LifeLineChartUnknownPlacementError("Something went wrong in the placement algorithm")
        elif len(strongly_connected_spouse_families) > 1:
            raise LifeLineChartUnknownPlacementError("Something went wrong in the placement algorithm")
        strongly_connected_parent_family = None
        strongly_connected_spouse_family = None
        if len(strongly_connected_parent_families) > 0:
            strongly_connected_parent_family = strongly_connected_parent_families[0]
        if len(strongly_connected_spouse_families) > 0:
            strongly_connected_spouse_family = strongly_connected_spouse_families[0]
        return strongly_connected_parent_family, strongly_connected_spouse_family

    @ancestor_chart_parent_family_placement.setter
    def ancestor_chart_parent_family_placement(self, gr_families):
        gr_parent_family, gr_spouse_family = gr_families
        if gr_parent_family is not None:
            self.__instances.connection_container['i'][self.g_id][gr_parent_family.g_id].append('strong_child')
            self.__instances.connection_container['f'][gr_parent_family.g_id][self.g_id].append('strong_child')
        if gr_spouse_family is not None:
            self.__instances.connection_container['i'][self.g_id][gr_spouse_family.g_id].append('strong_marriage')
            self.__instances.connection_container['f'][gr_spouse_family.g_id][self.g_id].append('strong_marriage')

    def get_name(self):
        return self.individual.get_name()

    def __get_birth_date(self):
        if self.individual.events['birth_or_christening']:
            return self.individual.events['birth_or_christening']['date'].date().strftime('%d.%m.%Y')
        else:
            return None
    birth_date = property(__get_birth_date)

    def __get_death_date(self):
        if self.individual.events['death_or_burial']:
            return self.individual.events['death_or_burial']['date'].date().strftime('%d.%m.%Y')
        else:
            return None
    death_date = property(__get_death_date)

    def get_position_dict(self, gr_family=None):
        if gr_family is None:
            return self._x_position
        else:
            return self._x_position.get(gr_family.g_id)

    def get_parent_positions(self):
        return OrderedDict([(k, v) for k, v in self._x_position.items() if v[3]])

    def get_spouse_positions(self):
        return OrderedDict([(k, v) for k, v in self._x_position.items() if not v[3]])

    def get_x_index(self, g_id):
        return self._x_position.get(g_id)[1]

    def get_other_family_connected_to_birth_position(self, gr_family):
        if gr_family is not None:
            g_id = gr_family.g_id
        else:
            g_id = None

        scpf, scsf = self.ancestor_chart_parent_family_placement

        # ancestor_placement_family = self.ancestor_placement_marriage

        # if scsf is None and scpf is None and ancestor_placement_family is not None:
        #     scsf = ancestor_placement_family

        other_g_id = "there is no connected family"
        scsf_g_id = scsf.g_id if scsf else None
        scpf_g_id = scpf.g_id if scpf else None
        if scsf or scpf:
            if g_id == scsf_g_id:
                # g_id is the strongyl connected spouse family
                other_g_id = scpf_g_id
            elif g_id == scpf_g_id:
                # g_id is the strongly connected parent family
                other_g_id = scsf_g_id

        return other_g_id

    def has_position_vector(self, family):
        if self._x_position is None:
            return False
        g_id = None
        if family is not None:
            g_id = family.g_id
        if g_id in self._x_position:
            return True
        return False

    def set_position_vector(self, x_position, gr_family, this_is_the_parent_family=False, overrule_ov=None):
        if gr_family:
            g_id = gr_family.g_id
            if gr_family.marriage:
                ov = gr_family.marriage['ordinal_value']
            else:
                ov = self.birth_date_ov
        else:
            g_id = None
            ov = self.birth_date_ov
        if overrule_ov:
            ov = overrule_ov
        if not self._x_position:
            self._x_position = OrderedDict()
        if g_id not in self._x_position:
            self._x_position[g_id] = (
                (ov, x_position, gr_family, this_is_the_parent_family))
        _x_position = OrderedDict()
        _x_position.update(OrderedDict(sorted(self._x_position.items(), key=lambda t: t[1][0])))
        self._x_position = _x_position

    x_position = property(get_position_dict, set_position_vector)

    @property
    def ancestor_placement_marriage(self):
        """
        The birth x position can be linked to a specific marriage. Ancestors will be placed above
        the birth x position.
        """
        if self.g_id not in self.__instances.connection_container['i']:
            return []
        gr_marriage_families = []
        for g_id, connections in self.__instances.connection_container['i'][self.g_id].items():
            if 'strong_marriage' in connections:
                gr_marriage_families.append(self.__instances[('f', g_id[1])].graphical_representations[g_id[0]])
        if len(gr_marriage_families) == 1:
            return gr_marriage_families[0]
        elif len(gr_marriage_families) > 1:
            raise LifeLineChartUnknownPlacementError("too many ancestor placement marriages " + str(gr_marriage_families) + " " + str(self))
        return None

    @ancestor_placement_marriage.setter
    def ancestor_placement_marriage(self, gr_marriage_family):
        """
        The birth x position can be linked to a specific marriage. Ancestors will be placed above
        the birth x position.

        Args:
            gr_marriage_family (GraphicalFamily): marriage where to place the ancestors
        """
        if gr_marriage_family:
            g_id = gr_marriage_family.g_id
        else:
            return
            g_id = None
        self.__instances.connection_container['i'][self.g_id][g_id].append('strong_marriage')
        self.__instances.connection_container['f'][g_id][self.g_id].append('strong_marriage')

    def is_cross_connection(self, gr_family_a, gr_family_b):
        """
        Returns true, if the two families do not share the same x position.

        Args:
            gr_family_a (GraphicalFamily): family a
            gr_family_b (GraphicalFamily): family b

        Returns:
            bool: is cross-connection
        """
        gr_family_g_id_a = gr_family_a.g_id if gr_family_a else None
        gr_family_g_id_b = gr_family_b.g_id if gr_family_b else None
        gr_family_pos_a = self._x_position.get(gr_family_g_id_a)
        gr_family_pos_b = self._x_position.get(gr_family_g_id_b)
        if gr_family_pos_a and gr_family_pos_b:
            if gr_family_pos_a[1] != gr_family_pos_b[1]:
                return True
            return False
        # this happens in descendant charts. spouses dont have a visible parent family.
        # raise LifeLineChartUnknownPlacementError("individual was not placed in requested family")
        return False

    def get_birth_event(self):
        return self.individual.events['birth_or_christening']

    def get_death_event(self):
        return self.individual.events['death_or_burial']

    @property
    def birth_date_ov(self):
        """
        Get the ordinal value of the birth (or christening or baptism) date.

        Returns:
            float: ordinal value of birth date
        """
        if self.__birth_date_ov is None:
            self.__birth_date_ov = self.individual.birth_date_ov
            return self.__birth_date_ov
        else:
            return self.__birth_date_ov

    @property
    def death_date_ov(self):
        """
        Get the ordinal value of the death (or burial) date.

        Returns:
            float: ordinal value of death date
        """
        if self.__death_date_ov is None:
            self.__death_date_ov = self.individual.death_date_ov
            return self.__death_date_ov
        else:
            return self.__death_date_ov

    @property
    def birth_label(self):
        return self.individual.birth_label  # + self.debug_label

    @property
    def death_label(self):
        string = self.individual.death_label
        if len(self._x_position) > 2 or len(self._x_position) == 2 and list(self._x_position.values())[0][1] != list(self._x_position.values())[1][1]:
            string += ' ' + self.individual.plain_name
        return string

    @property
    def children(self):
        return self.individual.children

    @property
    def visible_children(self):
        """
        Get the all visible children of this individual (all visible marriages).

        Returns:
            list: list of children individuals
        """
        gr_children = []
        for vm in self.visible_marriages:
            gr_children += vm.visible_children
        return gr_children

    def get_all_ancestors(self):
        """
        get all ancestors places over this individual

        Returns:
            list: list of ancestors
        """
        gr_ancestors = []
        strongly_connected_parent_family, strongly_connected_spouse_family = self.ancestor_chart_parent_family_placement
        if strongly_connected_parent_family:
            if strongly_connected_parent_family.gr_husb:
                gr_ancestors += [(strongly_connected_parent_family, strongly_connected_parent_family.gr_husb)]
                gr_ancestors += strongly_connected_parent_family.gr_husb.get_all_ancestors()
            gr_ancestors += [(strongly_connected_parent_family, c) for c in strongly_connected_parent_family.visible_children]
            if strongly_connected_parent_family.gr_wife:
                gr_ancestors += [(strongly_connected_parent_family, strongly_connected_parent_family.gr_wife)]
                gr_ancestors += strongly_connected_parent_family.gr_wife.get_all_ancestors()
        return gr_ancestors

    def get_all_descendants(self):
        gr_descendants = []
        for vm in self.visible_marriages:
            for vc in vm.visible_children:
                if not self.connected_parent_families or vm.descendant_chart_parent_family_placement == self.connected_parent_families[0]:
                    gr_descendants += [(vm, vc)]
                    gr_descendants += vc.get_all_descendants()
        return gr_descendants
